"""Provides custom callbacks for the Hugging Face Trainer class."""

from transformers import TrainerCallback
from pathlib import Path
import pandas as pd


class MetricsLoggerCallback(TrainerCallback):
    def _load_current_file(self, output_dir):
        output_dir = Path(output_dir)
        metrics_file = output_dir / "metrics.csv"

        return (
            pd.read_csv(metrics_file)
            if metrics_file.exists()
            else pd.DataFrame()
        )

    def on_log(self, args, state, control, logs=None, **kwargs):
        """
        Triggered during training logging (e.g., for training loss).
        """
        if "loss" in logs:  # Ensure training loss is available
            metrics_df = self._load_current_file(args.output_dir)

            # Append the epoch and the training loss.
            loss = logs["loss"]
            epoch = state.epoch

            metrics_df = pd.concat(
                [
                    metrics_df,
                    pd.DataFrame([{"epoch": epoch, "train_loss": loss}]),
                ],
                ignore_index=True,
            )

            # Save it back to file.
            metrics_df.to_csv(
                Path(args.output_dir) / "metrics.csv", index=False
            )

    def on_evaluate(self, args, state, control, metrics, **kwargs):
        """
        This is triggered at the end of evaluation (including at the end of each epoch).
        """
        # Get the current metrics file.
        metrics_df = self._load_current_file(args.output_dir)

        # Add metrics to the last row of the current metrics file.
        for k, v in metrics.items():
            metrics_df.at[metrics_df.index[-1], k] = v

        # Save the updated metrics file.
        metrics_df.to_csv(Path(args.output_dir) / "metrics.csv", index=False)



# -*- coding: utf-8 -*-
"""Tile functions.

This module contains functions used when working tiles.

"""
from geopandas import GeoDataFrame
from .image_utils import label_mask_to_polygons


def merge_tile_masks(
    tile_list: list, buffer: int = 1, background_label: int | None = 0
) -> GeoDataFrame:
    """
    Merge the tile masks into a single mask for a large image.

    Args:
        tile_list (list): List of three-lenght tuples containing:
            (1) fp or array for mask, (2) x-coordinate for tile, (3)
            y-coordinate for tile.
        buffer (int): Buffer to apply to the polygons before dissolving.
        background_label (int | None): Label value of the background class,
            which is ignored. Default is 0. If None then all labels are
            considered.

    Returns:
        GeoDataFrame: GeoDataFrame with the merged mask.

    """
    polygons_and_labels = []

    # Process each tile.
    for tile_info in tile_list:
        tile, x, y = tile_info

        # Process the mask by converting it to polygons.
        polygons_and_labels.extend(
            label_mask_to_polygons(
                tile, x_offset=x, y_offset=y, exclude_labels=background_label
            )
        )

    # Convert polygons and labels into a GeoDataFrame.
    gdf = GeoDataFrame(polygons_and_labels, columns=["geometry", "label"])

    # Apply a buffer to make edges touch.
    gdf["geometry"] = gdf["geometry"].buffer(buffer)

    # Dissolve the dataframe by the label.
    gdf = gdf.dissolve(by="label", as_index=False)

    # Remove the buffer.
    gdf["geometry"] = gdf["geometry"].buffer(-buffer)

    # Check that all polygons are valid.
    for i, row in gdf.iterrows():
        if not row["geometry"].is_valid:
            gdf.loc[i, "geometry"] = row["geometry"].buffer(0)

    return gdf



# Shadow imports.
from .imread import imread
from .imwrite import imwrite

# Version of the dsa-helpers package
__version__ = "1.1.5"

# To avoid slow downs, do not allow from dsa_helpers import * to import anything.
__all__ = []



from datasets import Dataset, Features, Image, Value
import pandas as pd
from .datasets import SegFormerSegmentationDataset


def dataset_generator(dataset):
    """Yield a dataset."""
    for item in dataset:
        yield item


def create_segformer_segmentation_dataset(
    df: pd.DataFrame | str,
    low_memory: bool = True,
    transforms=None,
    extras: bool = False,
    group: str | None = None,
) -> SegFormerSegmentationDataset:
    """Create a SegFormer segmentation dataset from a DataFrame.

    Args:
        df (pandas.DataFrame): Must have columns "fp" and "mask_fp".
            Additionally, it can have columns "x", "y". If these are not
            present, then they will be set to 0.
        low_memory (bool, optional): Whether to read the CSV file in low
            memory mode. Default is True.
        transforms: Function to add transforms to the images.
        group (str | None, optional): Additional column to pass to
            the model, if None it will be an empty string. Default
            is None.

    Returns:
        (SegFormerSegmentationDataset): A Dataset object to be used for
        HuggingFaces SegFormer model training.

    """
    if isinstance(df, str):
        df = pd.read_csv(df, low_memory=low_memory)

    dataset = SegFormerSegmentationDataset(df, group=group, extras=extras)

    if extras:
        features = Features(
            {
                "pixel_values": Image(),
                "label": Image(),
                "x": Value("int32"),
                "y": Value("int32"),
                "group": Value("string"),
            }
        )
    else:
        features = Features(
            {
                "pixel_values": Image(),
                "label": Image(),
            }
        )
    dataset = Dataset.from_generator(
        generator=lambda: dataset_generator(dataset), features=features
    )

    if transforms is not None:
        dataset.set_transform(transforms)

    return dataset



"""Script to develop functions for the library."""

from argparse import ArgumentParser
from dsa_helpers.girder_utils import (
    login,
    post_annotation,
    remove_overlapping_annotations,
)


def parse_args():
    parser = ArgumentParser()
    parser.add_argument(
        "--api-token", type=str, default=None, help="Girder API token"
    )
    parser.add_argument(
        "--api-url", type=str, default=None, help="Girder API URL"
    )
    return parser.parse_args()


def main(args):
    gc = login(args.api_url, api_key=args.api_token)

    # Get an annotation document.
    doc = gc.get("annotation/...")

    annotation = remove_overlapping_annotations(
        doc["annotation"], ["Gray Matter", "White Matter", "Leptomeninges"]
    )

    # Push the new doc.
    annotation["name"] = "temp"

    response = post_annotation(gc, "...", annotation)

    print(response)


if __name__ == "__main__":
    main(parse_args())



from torch.utils.data import Dataset
import pandas as pd
from PIL import Image
from colorama import Style, Fore


class SegFormerSegmentationDataset(Dataset):
    """DEPRECATED: see
    dsa_helpers.ml.segformer_semantic_segmentation.datasets.SegFormerSegmentationDataset

    PyTorch dataset class for semantic segmentation using HuggingFaces SegFormer
    model.

    """

    def __init__(self, df: pd.DataFrame):
        """Initiate an instance of the segmentation dataset.

        Args:
            df (pandas.DataFrame): A pandas DataFrame with columns "fp" and "mask_fp".

        """
        print(Fore.RED)
        print(
            "This is deprecated, please import from dsa_helpers.ml.segformer_semantic_segmentation.datasets"
        )
        print(Style.RESET_ALL)
        self.image_files = df["fp"].tolist()
        self.mask_files = df["mask_fp"].tolist()
        self.df = df  # store potential metadata

    def __len__(self):
        return len(self.image_files)

    def __getitem__(self, idx):
        # Get the filepath to the image and its mask.
        image_path = self.image_files[idx]
        mask_path = self.mask_files[idx]

        # Return image & mask as PIL images in a dictionary.
        return {
            "pixel_values": Image.open(image_path),
            "label": Image.open(mask_path),
        }



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "DSA Helpers"
copyright = "2024, Juan Carlos Vizcarra"
author = "Juan Carlos Vizcarra"
release = "00.00.01"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



"""
Utility functions for working with GeoPandas.

Functions:
- count_polygon_points: Count the total number of points in a Shapely polygon.
- total_points_in_gdf: Calculate the total number of points in all polygons in a GeoDataFrame.
- plot_gdf: Create a figure from a geopandas dataframe.
- remove_gdf_overlaps: Remove overlaps from a GeoDataFrame.
- draw_gdf_on_array: Draw a GeoDataFrame on an array.

"""

import geopandas as gpd
from shapely.geometry import Polygon
from rasterio.features import rasterize
import numpy as np
import matplotlib.patches as mpatches
import matplotlib.pyplot as plt


def count_polygon_points(polygon: Polygon) -> int:
    """
    Count the total number of points in a Shapely polygon, including
    exterior and interior rings.

    Args:
        polygon (shapely.geometry.Polygon): A Shapely polygon object.

    Returns:
        int: The total number of points in the polygon.

    """
    # Count exterior points.
    exterior_points = len(polygon.exterior.coords)

    # Count interior points (holes).
    interior_points = sum(
        len(interior.coords) for interior in polygon.interiors
    )

    # Total points.
    return exterior_points + interior_points


def total_points_in_gdf(gdf: gpd.GeoDataFrame) -> int:
    """
    Calculate the total number of points in all polygons in a GeoDataFrame.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame containing polygons.

    Returns:
        int: The total number of points in all polygons.

    """
    return int(gdf.geometry.apply(count_polygon_points).sum())


def plot_gdf(
    gdf: gpd.GeoDataFrame,
    figsize: int | tuple[int, int] | None = None,
    label_column: str = "group",
    label2color: dict | None = None,
) -> tuple[plt.Figure, plt.Axes]:
    """Create figure from a geopandas dataframe.

    Args:
        gdf (geopandas.GeoDataFrame): A GeoDataFrame containing
            polygons. Must have a 'label' column.
        figsize (int | tuple[int, int] | None): The size of the figure.
            If None, the figure size will be (5, 5).
        label_column (str): The column name of the label.
        label2color (dict | None): A dictionary mapping labels to
            colors. If None then a legend will not be included and the
            colors will be random. Colors should be an accepted
            matplotlib color.

    Returns:
        tuple[plt.Figure, plt.Axes]: The figure and axes objects.

    """
    if figsize is None:
        figsize = (5, 5)
    elif isinstance(figsize, int):
        figsize = (figsize, figsize)

    gdf = gdf.copy()

    # Blow up any multi-polygons.
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    if label2color is not None:
        gdf["color"] = gdf[label_column].map(label2color)

        # Remove rows where color is nan.
        gdf = gdf.dropna(subset=["color"])

    # Create the plot
    fig, ax = plt.subplots(figsize=figsize)
    gdf.plot(color=gdf["color"], ax=ax)

    if label2color is not None:
        # Add legend.
        handles = [
            mpatches.Patch(color=color, label=key)
            for key, color in label2color.items()
        ]

        ax.legend(
            handles=handles,
            title="Legend",
            loc="upper left",
            bbox_to_anchor=(
                1.02,
                1,
            ),  # x > 1 pushes it to the right, y = 1 keeps it top-aligned
            borderaxespad=0,
        )

    return fig, ax


def remove_gdf_overlaps(
    gdf: gpd.GeoDataFrame,
    columns: str | list[str],
    ascending: bool | list[bool] = True,
    is_notebook: bool = False,
) -> gpd.GeoDataFrame:
    """Remove overlaps from a GeoDataFrame.

    Args:
        gdf (geopandas.GeoDataFrame): The GeoDataFrame to remove
            overlaps from.
        columns (str | list[str]): The columns to sort dataframe by. The
            order determines the order of removal overlapping polygon
            regions.
        ascending (bool): Whether to remove overlaps in ascending order.
        is_notebook (bool): Switch which tqdm bar to use. Default is
            False, which uses the terminal progress bar.

    Returns:
        gpd.GeoDataFrame: The GeoDataFrame with overlaps removed.

    """
    if isinstance(columns, str):
        columns = [columns]

    if isinstance(ascending, bool):
        ascending = [ascending] * len(columns)

    # Remove multi-polygons.
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    # Sort the GeoDataFrame by the columns and ascending order.
    gdf = gdf.sort_values(by=columns, ascending=ascending).reset_index(
        drop=True
    )

    if is_notebook:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm

    # Loop through each row in the GeoDataFrame.
    for i, row in tqdm(gdf.iterrows(), total=len(gdf)):
        # Check overlap with all following rows.
        other_gdf = gdf[i + 1 :]

        if len(other_gdf):
            # Get the rows that overlap.
            overlap = row.geometry.overlaps(other_gdf.geometry)

            # Get the rows that overlap.
            other_rows = other_gdf[overlap]

            other_gdf.loc[:, "geometry"] = other_gdf.geometry.difference(
                row.geometry
            )

    # This could have created multiple polygons, explode them.
    gdf = gdf.explode(index_parts=False).reset_index(drop=True)

    return gdf


def draw_gdf_on_array(gdf, shape, id_column="idx", default_value: int = 0):
    """Use the polygons in a geopandas dataframe to draw on a passed
    numpy array.

    Args:
        gdf (geopandas.GeoDataFrame): The GeoDataFrame to draw on the
            array.
        shape (tuple): Height and width of the array to draw on.
        id_column (str): The column on the dataframe to use for the
            value of each polygon drawn. The values in this column
            must be integers.
        default_value (int): The default value of the array, this will
            the values returned for pixels that are not drawn on.

    Returns:
        numpy.ndarray: The array with the polygons drawn on it.

    """
    # Create the array.
    array = np.ones(shape, dtype=np.uint8) * default_value

    # Draw each polygon with its corresponding label
    for _, row in gdf.iterrows():
        # Create a list of (geometry, value) pairs
        shapes = [(row.geometry, row[id_column])]

        # Rasterize the shapes
        burned = rasterize(
            shapes=shapes,
            out_shape=shape,
            fill=default_value,
            default_value=default_value,
        )

        # Update the array where the polygon was burned
        array = np.where(burned > 0, burned, array)

    return array



from transformers import SegformerImageProcessor
from torchvision.transforms import ColorJitter


def val_transforms_extras(example_batch):
    """Default transforms for validation images with extras."""
    processor = SegformerImageProcessor()

    images = [x for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]
    inputs = processor(images, labels)
    inputs["x"] = example_batch["x"]
    inputs["y"] = example_batch["y"]
    inputs["group"] = example_batch["group"]

    return inputs


def val_transforms(example_batch):
    """Default transforms for validation images."""
    processor = SegformerImageProcessor()

    images = [x for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]

    inputs = processor(images, labels)

    return inputs


def train_transforms_extras(example_batch):
    """Default transforms for training images with extras."""
    processor = (
        SegformerImageProcessor()
    )  # required for using SegFormer model.
    jitter = ColorJitter(
        brightness=0.25, contrast=0.25, saturation=0.25, hue=0.1
    )

    images = [jitter(x) for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]
    inputs = processor(images, labels)
    inputs["x"] = example_batch["x"]
    inputs["y"] = example_batch["y"]
    inputs["group"] = example_batch["group"]

    return inputs


def train_transforms(example_batch):
    """Default transforms for training images."""

    processor = SegformerImageProcessor()
    jitter = ColorJitter(
        brightness=0.25, contrast=0.25, saturation=0.25, hue=0.1
    )

    images = [jitter(x) for x in example_batch["pixel_values"]]
    labels = [x for x in example_batch["label"]]

    inputs = processor(images, labels)

    return inputs
